from direct.fsm import FSM
import math

#-------------------------------------
# FSM that manages walking animations.
class WalkAnimFSM(FSM.FSM):
    def __init__(self, actor, walkRate, reverseWalkRate, turnWalkRate):
        FSM.FSM.__init__(self, "WalkAnimFSM_" + str(id(self)))
        self.actor = actor
        self.walkRate = walkRate
        self.reverseWalkRate = reverseWalkRate
        self.turnWalkRate = turnWalkRate
    def enterWalk(self):
        curAnim = self.actor.getCurrentAnim()
        if (self.oldState != "Walk" or curAnim != "walk") and curAnim != "jump":
            self.actor.setPlayRate(self.walkRate, "walk")
            self.actor.loop("walk", restart = 0)
    def exitWalk(self):
        pass
    def enterReverseWalk(self):
        curAnim = self.actor.getCurrentAnim()
        if (self.oldState != "ReverseWalk" or curAnim != "walk") and curAnim != "jump":
            self.actor.setPlayRate(-self.reverseWalkRate, "walk")
            self.actor.loop("walk", restart = 0)
    def exitReverseWalk(self):
        pass
    def enterTurnWalk(self):
        curAnim = self.actor.getCurrentAnim()
        if (self.oldState != "TurnWalk" or curAnim != "walk") and curAnim != "jump":
            self.actor.setPlayRate(self.turnWalkRate, "walk")
            self.actor.loop("walk", restart = 0)
    def exitTurnWalk(self):
        pass
    def enterStill(self):
        if self.actor.getCurrentAnim() != "jump":
            self.actor.stop()
    def exitStill(self):
        pass
    ##end def
##end class

#---------------------------------------------------------------
# Plays a sound, altering the volume to take range into account.
def playRangedSound(sound, x1, y1, x2, y2, maxRange):
    xDiff = x1 - x2
    yDiff = y1 - y2
    distSquared = xDiff*xDiff + yDiff*yDiff
    if distSquared <= maxRange*maxRange:
        dist = math.sqrt(distSquared)
        volume = 1 - (dist / maxRange)
        sound.setVolume(volume)
        sound.play()
    ##end if
##end def

#----------------------------------------------------------------------
# Recursively sets a tag to the specified node and all of its chidlren.
def recursiveSetTag(node, tag, value):
    node.setTag(tag, value)
    for child in node.getChildren():
        recursiveSetTag(child, tag, value)
##end def

#-------------------------------------------------------------------------
# Recursively sets the name of the specified node and all of its children.
def recursiveSetName(node, name):
    node.setName(name)
    for child in node.getChildren():
        recursiveSetName(child, name)
##end def